#import "GraphGLView.h"


@implementation GraphGLView

- (id)initWithFrame:(NSRect)frame
{
    GLuint attribs[] = {
        NSOpenGLPFANoRecovery,
        NSOpenGLPFAWindow,
        NSOpenGLPFAAccelerated,
        NSOpenGLPFADoubleBuffer,
        NSOpenGLPFAColorSize, 24,
        NSOpenGLPFAAlphaSize, 8,
        NSOpenGLPFADepthSize, 24,
        NSOpenGLPFAStencilSize, 8,
        NSOpenGLPFAAccumSize, 0,
        0
    };
    
    NSOpenGLPixelFormat* fmt = [[NSOpenGLPixelFormat alloc] initWithAttributes: (NSOpenGLPixelFormatAttribute*) attribs]; 
    
    if (!fmt) NSLog(@"No OpenGL pixel format");
    
    return self = [super initWithFrame:frame pixelFormat: [fmt autorelease]];
}

- (void)awakeFromNib
{
    m_state = OFF;
	m_cycleType = NO_CYCLING;
	m_drawGridLines = TRUE;
	m_drawTempPlot = TRUE;
	m_xAxisType = TIME_AXIS;
	
	m_currentZeroPoint = 1.25; // about right for a starting point..
	
	m_ampHourIn = m_ampHourOut = m_wattHourIn = m_wattHourOut = 0;
    
    counter = 0;
	
	for (int n=0; n<10; n++)
	{
		voltageSample[n] = 0;
		currentSample[n] = 0;
		tempSample[n] = 0;
	}
	c = 0;
    
	NSString* temp = [[[NSBundle mainBundle] resourcePath] stringByAppendingString:@"/Pop.aif"];
    m_changeStateSound = [[NSSound alloc] initWithContentsOfFile:temp byReference:YES];
    
    temp = [[[NSBundle mainBundle] resourcePath] stringByAppendingString:@"/ChaChing.aif"];
    m_cycleFinishedSound = [[NSSound alloc] initWithContentsOfFile:temp byReference:YES];
    
    m_timer = [[NSTimer scheduledTimerWithTimeInterval:0.05
                                               target:self
                                             selector:@selector(tick)
                                             userInfo:nil
                                              repeats:YES] retain];
	
	m_timer = [[NSTimer scheduledTimerWithTimeInterval:1
												target:self
											  selector:@selector(redraw)
											  userInfo:nil
											   repeats:YES] retain];
	
	m_accurateTimer = [[Timer alloc] init];
	[m_accurateTimer reset];
}


- (void)tick
{
	c++;
	voltageSample[c%10] = [m_controller getAnalogueIn:3] * 2.0;
	currentSample[c%10] = ([m_controller getAnalogueIn:2] - m_currentZeroPoint) * 52; //([m_controller getAnalogueIn:2] * 2.0 - voltageSample[c%10] - 0.008) * 2.67;
	tempSample[c%10] = 3 + [m_controller getAnalogueIn:4]*20;
	
	//if (currentSample[c%10] > 0) currentSample[c%10] /= 1.16;
	
	if (m_state == OFF)
	{
		if (currentSample[c%10] > 0) m_currentZeroPoint *= 1.0001;
		if (currentSample[c%10] < 0) m_currentZeroPoint /= 1.0001;
	}
	
	voltage = 0; current = 0; temp = 0;
	for (int n=0; n<10; n++)
	{
		voltage += voltageSample[c%10];
		current += currentSample[c%10];
		temp += tempSample[c%10];
	}
	voltage /= 10;
	current /= 10;
	temp /= 10;
	
	[m_voltageField setFloatValue:voltage];
	[m_currentField setFloatValue:current];
	[m_temperatureField setFloatValue:temp];
	
	// State switching
	// Update states
	if (voltage > 3.6 && current < 0.3 && m_state == CHARGING)
	{
		if (m_ampHourIn > 0 && m_wattHourIn > 0 && m_ampHourOut > 0 && m_wattHourOut > 0 && m_cycleType != NO_CYCLING)
		{
			NSString* string = [[m_cycleLog string] stringByAppendingString:[NSString stringWithFormat:@"%.3f, %.3f, %.3f, %.3f\n",
				m_ampHourIn, m_wattHourIn, m_ampHourOut, m_wattHourOut]];
			[m_cycleLog setString:string];
		}
		
		if (m_cycleType == NO_CYCLING)
			[self stop:self];
		else
			[self discharge:self];
	}
	else if (voltage < 2.0 && m_state == DISCHARGING)
	{
		if (m_cycleType == CONTINUOUS_CYCLING)
		{
			[self charge:self];
			[self clearGraph:self];
		}
		else
			[self stop:self];
	}
	
	if (m_state != OFF && counter < 99999 && [m_accurateTimer getElapsedSeconds] > 1)
    {
		counter++;
        m_voltages[counter] = voltage;
        m_currents[counter] = current;
		m_temps[counter] = temp;
		
		double elapsedTime = [m_accurateTimer getElapsedSeconds];
		[m_accurateTimer addOffset:-1];
		
		if (current > 0.1) // charging
		{
			m_ampHourIn += current * elapsedTime / 3600;
			m_wattHourIn += voltage * current * elapsedTime / 3600;
		}
		else if (current < -0.1) // discharging
		{
			m_ampHourOut -= current * elapsedTime / 3600;
			m_wattHourOut -= voltage * current * elapsedTime / 3600;
		}
		
		NSString* string = [[m_log string] stringByAppendingString:[NSString stringWithFormat:@"%.3f, %.3f\n", voltage, current]];
		
		[m_log setString:string];
		
    }
	
	[m_ampHourInField setFloatValue:m_ampHourIn];
	[m_wattHourInField setFloatValue:m_wattHourIn];
	[m_ampHourOutField setFloatValue:m_ampHourOut];
	[m_wattHourOutField setFloatValue:m_wattHourOut];
	
	
    // Set relay drivers
    BOOL chargeRelayState = NO;	
	if (m_state == CHARGING) chargeRelayState = YES;
    [m_controller setDigitalOut:chargeRelayState channel:1];
    
    BOOL dischargeRelayState = NO;
    if (m_state == DISCHARGING) dischargeRelayState = YES;
    [m_controller setDigitalOut:dischargeRelayState channel:0];
}

- (void)redraw
{
    [self setNeedsDisplay:YES];
}

- (void)drawRect:(NSRect)rect
{
    
    glViewport(0, 0, rect.size.width, rect.size.height);
    
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    
    glOrtho(-0.01, 1.01, -0.05, 5.05, 0, 1);
    
    glMatrixMode(GL_MODELVIEW);
    
    glClearColor(0, 0, 0, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    //glEnable(GL_DEPTH_TEST);
        
    
	if (m_drawGridLines)
	{
		glBegin(GL_LINES);
		glColor3f(0, 0.2, 0);
    
		if (m_xAxisType == TIME_AXIS)
		{
			float minutes = (float)counter/60;
			for (double x=1; x<minutes; x++)
			{
				glVertex2f(x*60/counter, 0);
				glVertex2f(x*60/counter, 5);
			}
		}
		else
		{
			float charge = m_ampHourIn+m_ampHourOut;
			for (double x=0; x<charge; x+= 0.01)
			{
				glVertex2f(x/charge, 0);
				glVertex2f(x/charge, 5);
			}
		}
		for (double y=0; y<5; y+=0.1)
		{
			glVertex2f(0, y);
			glVertex2f(1, y);
		}
		glColor3f(0, 0.5, 0);
		for (double y=0; y<5; y+=1)
		{
			glVertex2f(0, y);
			glVertex2f(1, y);
		}
		glEnd();
	}
	
    glBegin(GL_LINE_LOOP);
    glColor3f(0, 0.8, 0);
    glVertex2f(0, 0);
    glVertex2f(1, 0);
    glVertex2f(1, 5);
    glVertex2f(0, 5);
    glEnd();
    
	double x, ampHour = 0;
	
	glBegin(GL_LINE_STRIP);
    glColor3f(1, 1, 1);
    for (int n=0; n<=counter; n++)
    {
		// work out x position
		ampHour += fabs(m_currents[n])/3600;
		if (m_xAxisType == TIME_AXIS)
			x = (double)n/counter;
		else
			x = ampHour/(m_ampHourIn + m_ampHourOut);
		
		glVertex2f(x, m_voltages[n]);
    
	}
	glEnd();
    
    glBegin(GL_LINE_STRIP);
    glColor3f(1, 0.5, 0);
	ampHour = 0;
	for (int n=0; n<=counter; n++)
	{
		// work out x position
		ampHour += fabs(m_currents[n])/3600;
		if (m_xAxisType == TIME_AXIS)
			x = (double)n/counter;
		else
			x = ampHour/(m_ampHourIn + m_ampHourOut);
		
		float v = m_currents[n];
		if (v > 0)
			glColor3f(1.0, 0.5, 0);
		else
			glColor3f(0, 0.5, 1.0);
		
		glVertex2f(x, fabs(v)/1.0); // sometimes need to scale current differently
	}
    glEnd();
	
	if (m_drawTempPlot)
	{
		glBegin(GL_LINE_STRIP);
		glColor3f(1, 0, 0);
		ampHour = 0;
		for (int n=0; n<=counter; n++)
		{
			// work out x position
			ampHour += fabs(m_currents[n])/3600;
			if (m_xAxisType == TIME_AXIS)
				x = (double)n/counter;
			else
				x = ampHour/(m_ampHourIn + m_ampHourOut);
			
			glVertex2f(x, m_temps[n]/10);
		}
		glEnd();
	}
    [[self openGLContext] flushBuffer];
}

- (IBAction)charge:(id)sender
{
    m_state = CHARGING;
	[m_accurateTimer reset];
	[m_changeStateSound play];
	[m_statusField setStringValue:@"CHARGING"];
}

- (IBAction)discharge:(id)sender
{
	m_state = DISCHARGING;
	[m_accurateTimer reset];
	[m_changeStateSound play];
	[m_statusField setStringValue:@"DISCHARGING"];
}

- (IBAction)stop:(id)sender
{
    m_state = OFF;
	[m_cycleFinishedSound play];
	[m_statusField setStringValue:@"OFF"];
}

- (IBAction)clearGraph:(id)sender
{
	counter = -1;
	[m_log setString:@""];
	[m_cycleLog setString:@""];
	m_ampHourIn = m_ampHourOut = m_wattHourIn = m_wattHourOut = 0;
}

- (IBAction)setCycleType:(id)sender
{
	m_cycleType = [sender tag];
	[m_noCyclingMenuItem setState:NSOffState];
	[m_singleCycleMenuItem setState:NSOffState];
	[m_continuousCyclingMenuItem setState:NSOffState];
	
	[sender setState:NSOnState];
}

- (IBAction)toggleGridLines:(id)sender
{
	m_drawGridLines = !m_drawGridLines;
}

- (IBAction)toggleTempPlot:(id)sender
{
	m_drawTempPlot = !m_drawTempPlot;
}

- (IBAction)setXAxisType:(id)sender
{
	m_xAxisType = [sender tag];
}

@end
